/* Copyright (c) INRIA and Microsoft Corporation. All rights reserved.
   Licensed under the Apache 2.0 and MIT Licenses. */

#ifndef KRML_HEADER_ENDIAN_H
#define KRML_HEADER_ENDIAN_H

#ifdef __GNUC__
#warning "c_endianness.h is deprecated, include lowstar_endianness.h instead"
#endif

#include "lowstar_endianness.h"

#endif /* KRML_HEADER_ENDIAN_H */
